/*
 * @(#)image.c
 *
 * create_images() Copyright 2000  Stephen Montgomery-Smith
 *
 * Copyright 1993 - 2008  David A. Bagley, bagleyd@tux.org
 *
 * Taken from GENERIC X-BASED TETRIS
 *
 * Copyright 1992 - 1995  Q. Alex Zhao, azhao@cc.gatech.edu
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

/* images for Tetris & Welltris */

#include <stdlib.h>
#include <stdio.h>
#include "xwin.h"
#include "image.h"

#define ARR(i,j,p,s) (((i)<0||(i)>=s||(j)<0||(j)>=s)?-2:(p[j][i]>0))

#define ROUND8(n) ((((n)+7)/8)*8)

/* A set bit indicates that an edge or corner is required. */
#define IS_LEFT(n) ((n)&LEFT)
#define IS_DOWN(n) ((n)&DOWN)
#define IS_RIGHT(n) ((n)&RIGHT)
#define IS_UP(n) ((n)&UP)
#define IS_LEFT_DOWN(n) ((n)&LEFT_DOWN)
#define IS_RIGHT_DOWN(n) ((n)&RIGHT_DOWN)
#define IS_RIGHT_UP(n) ((n)&RIGHT_UP)
#define IS_LEFT_UP(n) ((n)&LEFT_UP)

/* Defines to access the bitmaps. */
#ifdef WINVER
#define SETBIT(x,y) SetPixel(monoDC,x,y,RGB(255,255,255))
#define RESBIT(x,y) SetPixel(monoDC,x,y,RGB(0,0,0))
#else
#define BITNO(x,y) ((x)+(y)*ROUND8(boxsize))
#define SETBIT(x,y) data[BITNO(x,y)/8]|=1<<(BITNO(x,y)%8)
#define RESBIT(x,y) data[BITNO(x,y)/8]&=~(1<<(BITNO(x,y)%8))
#endif
#define TWOTHIRDSBIT(x,y) {if ((x+y+2)%3) SETBIT(x,y); else RESBIT(x,y);}
#define HALFBIT(x,y) {if ((x+y)%2) SETBIT(x,y); else RESBIT(x,y);}
#define THIRDBIT(x,y) {if (!((x+y+2)%3)) SETBIT(x,y); else RESBIT(x,y);}
#define THREEQUARTERSBIT(x,y) \
  {if (((y&1)==1)||(((x+2+y/2+1)&1)==1)) SETBIT(x,y); else RESBIT(x,y);}
#define NOTHALFBIT(x,y) {if ((x+y)%2) RESBIT(x,y); else SETBIT(x,y);}

/* Parameters for bitmaps. */
#define G ((boxsize/45)+1)         /* 1/2 of gap between pentominoes. */
#define T ((boxsize<=12)?1:(G*2))  /* Thickness of walls of pentominoes. */
#define R ((boxsize<=12)?1:(G*6))  /* Amount of rounding. */
#define RT ((boxsize<=12)?1:(G*3)) /* Thickness of wall of rounded parts.
                                  Here 3 is an approximation to 2 sqrt(2).  */
#define RR 0   /* Roof ridge thickness */

#ifdef WINVER
HBITMAP images[2][BITMAPS];
#else
XImage *images[2][BITMAPS];
#endif

/* A list of those bitmaps we need to create to display any polyomino. */
static int images_needed[] =
{
/* This needed for mononimo*/
	LEFT|RIGHT|UP|DOWN|LEFT_UP|LEFT_DOWN|RIGHT_UP|RIGHT_DOWN,

	LEFT_UP|LEFT_DOWN|RIGHT_UP|RIGHT_DOWN,

	LEFT|RIGHT_UP|RIGHT_DOWN,
	RIGHT|LEFT_UP|LEFT_DOWN,
	UP|LEFT_DOWN|RIGHT_DOWN,
	DOWN|LEFT_UP|RIGHT_UP,
	LEFT|RIGHT_UP,
	RIGHT|LEFT_UP,
	UP|LEFT_DOWN,
	DOWN|LEFT_UP,
	LEFT|RIGHT_DOWN,
	RIGHT|LEFT_DOWN,
	UP|RIGHT_DOWN,
	DOWN|RIGHT_UP,

	/* These needed for hexonimoes*/
	LEFT,
	RIGHT,
	UP,
	DOWN,
	LEFT_DOWN|RIGHT_UP|RIGHT_DOWN,
	LEFT_UP|RIGHT_UP|RIGHT_DOWN,
	LEFT_UP|LEFT_DOWN|RIGHT_DOWN,
	LEFT_UP|LEFT_DOWN|RIGHT_UP,

	LEFT|UP|RIGHT_DOWN,
	LEFT|DOWN|RIGHT_UP,
	RIGHT|UP|LEFT_DOWN,
	RIGHT|DOWN|LEFT_UP,
	LEFT|UP,
	LEFT|DOWN,
	RIGHT|UP,
	RIGHT|DOWN,

	LEFT|RIGHT,
	UP|DOWN,

	RIGHT|UP|DOWN,
	LEFT|UP|DOWN,
	LEFT|RIGHT|DOWN,
	LEFT|RIGHT|UP,

	-1
};

static char
image_needed(int n)
{
	int i;

	for (i = 0; images_needed[i] != -1; i++)
		if (n == images_needed[i])
			return 1;
	return 0;
}

static char
create_an_image(
#ifndef WINVER
		Display *display, Visual *visual,
#endif
		int n, char use3D, int next, int size)
{
	Boolean bgDark = True;
#ifdef WINVER
	int x, y, boxsize;
	HDC monoDC;
	HBITMAP hMask, hOldBitmap;

	boxsize = size;
	monoDC = CreateCompatibleDC(0);
	hMask = CreateCompatibleBitmap(monoDC, boxsize, boxsize);
	hOldBitmap = SelectObject(monoDC, hMask);
#else
	unsigned int x, y, boxsize;
	unsigned char * data;

	boxsize = (unsigned int) size;
	if ((data = (unsigned char *) calloc((boxsize*ROUND8(boxsize)/8),
			sizeof(char))) == NULL) {
		return 0; /* no biggie, will not use images */
	}
#endif
	for (y=0;y<boxsize;y++)
		for (x=0;x<boxsize;x++) {
			if (!use3D) {
#ifdef SMALL_BELLYBUTTON
				if (x >= boxsize / 2 - 1 && x <= boxsize / 2 &&
				    y >= boxsize / 2 - 1 && y <= boxsize / 2) {
					NOTHALFBIT(x,y);
				} else
#endif
				{
					HALFBIT(x,y);
				}
			} else if ((x>=y && x<=boxsize-y-1 && IS_UP(n)) ||
					(x<=y && x<=boxsize-y-1 &&
					y<boxsize/2 && !IS_LEFT(n)) ||
					(x>=y && x>=boxsize-y-1 &&
					y<boxsize/2 && !IS_RIGHT(n))) {
				if (bgDark) {
					SETBIT(x,y);
				} else {
					THIRDBIT(x,y);
				}
			} else if ((x<=y && x<=boxsize-y-1 && IS_LEFT(n)) ||
					(x>=y && x<=boxsize-y-1 &&
					x<boxsize/2 && !IS_UP(n)) ||
					(x<=y && x>=boxsize-y-1 &&
					x<boxsize/2 && !IS_DOWN(n))) {
				if (bgDark) {
					TWOTHIRDSBIT(x,y);
				} else {
					HALFBIT(x,y);
				}
			} else if ((x>=y && x>=boxsize-y-1 && IS_RIGHT(n)) ||
					(x>=y && x<=boxsize-y-1 &&
					x>=boxsize/2 && !IS_UP(n)) ||
					(x<=y && x>=boxsize-y-1 &&
					x>=boxsize/2 && !IS_DOWN(n))) {
				if (bgDark) {
					HALFBIT(x,y);
				} else {
					TWOTHIRDSBIT(x,y);
				}
			} else if ((x<=y && x>=boxsize-y-1 && IS_DOWN(n)) ||
					(x<=y && x<=boxsize-y-1 &&
					y>=boxsize/2 && !IS_LEFT(n)) ||
					(x>=y && x>=boxsize-y-1 &&
					y>=boxsize/2 && !IS_RIGHT(n))) {
				if (bgDark) {
					THIRDBIT(x,y);
				} else {
					SETBIT(x,y);
				}
			}
		}

	if (IS_LEFT(n))
		for (y=0;y<boxsize;y++)
			for (x=G;x<G+T;x++) {
				SETBIT(x,y);
			}
	if (IS_RIGHT(n))
		for (y=0;y<boxsize;y++)
			for (x=G;x<G+T;x++) {
				SETBIT(boxsize-1-x,y);
			}
	if (IS_UP(n))
		for (x=0;x<boxsize;x++)
			for (y=G;y<G+T;y++) {
				SETBIT(x,y);
			}
	if (IS_DOWN(n))
		for (x=0;x<boxsize;x++)
			for (y=G;y<G+T;y++) {
				SETBIT(x,boxsize-1-y);
			}
	if (IS_LEFT(n))
		for (y=0;y<boxsize;y++)
			for (x=0;x<G;x++) {
				RESBIT(x,y);
			}
	if (IS_RIGHT(n))
		for (y=0;y<boxsize;y++)
			for (x=0;x<G;x++) {
				RESBIT(boxsize-1-x,y);
			}
	if (IS_UP(n))
		for (x=0;x<boxsize;x++)
			for (y=0;y<G;y++) {
				RESBIT(x,y);
			}
	if (IS_DOWN(n))
		for (x=0;x<boxsize;x++)
			for (y=0;y<G;y++) {
				RESBIT(x,boxsize-1-y);
			}

	if (IS_LEFT(n) && IS_UP(n))
		for (x=G;x<=G+R;x++)
			for (y=G;y<=R+2*G-x;y++) {
				if (x+y>R+2*G-RT) {
					SETBIT(x,y);
				} else {
					RESBIT(x,y);
				}
			}
	if (IS_LEFT(n) && IS_DOWN(n))
		for (x=G;x<=G+R;x++)
			for (y=G;y<=R+2*G-x;y++) {
				if (x+y>R+2*G-RT) {
					SETBIT(x,boxsize-1-y);
				} else {
					RESBIT(x,boxsize-1-y);
				}
			}
	if (IS_RIGHT(n) && IS_UP(n))
		for (x=G;x<=G+R;x++)
			for (y=G;y<=R+2*G-x;y++) {
				if (x+y>R+2*G-RT) {
					SETBIT(boxsize-1-x,y);
				} else {
					RESBIT(boxsize-1-x,y);
				}
			}
	if (IS_RIGHT(n) && IS_DOWN(n))
		for (x=G;x<=G+R;x++)
			for (y=G;y<=R+2*G-x;y++) {
				if (x+y>R+2*G-RT) {
					SETBIT(boxsize-1-x,boxsize-1-y);
				} else {
					RESBIT(boxsize-1-x,boxsize-1-y);
				}
			}

	if (!IS_LEFT(n) && !IS_UP(n) && IS_LEFT_UP(n)) {
		for (x=0;x<G;x++)
			for(y=0;y<G;y++) {
				RESBIT(x,y);
			}
		for (x=G;x<G+T;x++)
			for(y=0;y<G;y++) {
				SETBIT(x,y);
			}
		for (x=0;x<G+T;x++)
			for(y=G;y<G+T;y++) {
				SETBIT(x,y);
			}
	}
	if (!IS_LEFT(n) && !IS_DOWN(n) && IS_LEFT_DOWN(n)) {
		for (x=0;x<G;x++)
			for(y=0;y<G;y++) {
				RESBIT(x,boxsize-1-y);
			}
		for (x=G;x<G+T;x++)
			for(y=0;y<G;y++) {
				SETBIT(x,boxsize-1-y);
			}
		for (x=0;x<G+T;x++)
			for(y=G;y<G+T;y++) {
				SETBIT(x,boxsize-1-y);
			}
	}
	if (!IS_RIGHT(n) && !IS_UP(n) && IS_RIGHT_UP(n)) {
		for (x=0;x<G;x++)
			for(y=0;y<G;y++) {
				RESBIT(boxsize-1-x,y);
			}
		for (x=G;x<G+T;x++)
			for(y=0;y<G;y++) {
				SETBIT(boxsize-1-x,y);
			}
		for (x=0;x<G+T;x++)
			for(y=G;y<G+T;y++) {
				SETBIT(boxsize-1-x,y);
			}
	}
	if (!IS_RIGHT(n) && !IS_DOWN(n) && IS_RIGHT_DOWN(n)) {
		for (x=0;x<G;x++)
			for(y=0;y<G;y++) {
				RESBIT(boxsize-1-x,boxsize-1-y);
			}
		for (x=G;x<G+T;x++)
			for(y=0;y<G;y++) {
				SETBIT(boxsize-1-x,boxsize-1-y);
			}
		for (x=0;x<G+T;x++)
			for(y=G;y<G+T;y++) {
				SETBIT(boxsize-1-x,boxsize-1-y);
			}
	}

#ifdef LARGE_BELLYBUTTON
	if (!use3D) {
		if (!IS_LEFT(n) && !IS_UP(n) && !IS_LEFT_UP(n))
			for (x=0;x<G+T;x++)
				for(y=0;y<G+T;y++) {
					SETBIT(x,y);
				}
		if (!IS_LEFT(n) && !IS_DOWN(n) && !IS_LEFT_DOWN(n))
			for (x=0;x<G+T;x++)
				for(y=boxsize-G-T;y<boxsize;y++) {
					SETBIT(x,y);
				}
		if (!IS_RIGHT(n) && !IS_UP(n) && !IS_RIGHT_UP(n))
			for (x=boxsize-G-T;x<boxsize;x++)
				for(y=0;y<G+T;y++) {
					SETBIT(x,y);
				}
		if (!IS_RIGHT(n) && !IS_DOWN(n) && !IS_RIGHT_DOWN(n))
			for (x=boxsize-G-T;x<boxsize;x++)
				for(y=boxsize-G-T;y<boxsize;y++) {
					SETBIT(x,y);
				}
	} else
#else
	if (use3D)
#endif
	{
		if (!IS_LEFT(n) && !IS_UP(n) && !IS_LEFT_UP(n))
			for (x=0;x<boxsize/2-RR;x++)
				for(y=0;y<boxsize/2-RR;y++) {
					THREEQUARTERSBIT(x,y);
				}
		if (!IS_LEFT(n) && !IS_DOWN(n) && !IS_LEFT_DOWN(n))
			for (x=0;x<boxsize/2-RR;x++)
				for(y=boxsize/2+RR;y<boxsize;y++) {
					THREEQUARTERSBIT(x,y);
				}
		if (!IS_RIGHT(n) && !IS_UP(n) && !IS_RIGHT_UP(n))
			for (x=boxsize/2+RR;x<boxsize;x++)
				for(y=0;y<boxsize/2-RR;y++) {
					THREEQUARTERSBIT(x,y);
				}
		if (!IS_RIGHT(n) && !IS_DOWN(n) && !IS_RIGHT_DOWN(n))
			for (x=boxsize/2+RR;x<boxsize;x++)
				for(y=boxsize/2+RR;y<boxsize;y++) {
					THREEQUARTERSBIT(x,y);
				}
	}
#ifdef WINVER
	images[next][n] = hMask;
	(void) SelectObject(monoDC, hOldBitmap);
	DeleteDC(monoDC);
#else
	if ((images[next][n] = XCreateImage(display, visual,
			1, XYBitmap, 0, (char *) data, boxsize, boxsize,
			8, 0)) == None) {
		free(data);
		return 0;
	}
	images[next][n]->byte_order = MSBFirst;
	images[next][n]->bitmap_unit = 8;
	images[next][n]->bitmap_bit_order = LSBFirst;
#endif
	return 1;
}

char
create_images(
#ifndef WINVER
		Display *display, Visual *visual,
#endif
		char use3D, int next, int size)
{
	int n;

	if (!create_an_image(
#ifndef WINVER
			display, visual,
#endif
			BITMAPS - 1, use3D, next, size))
		return 0;
	for (n = 0; n < BITMAPS - 1; n++) {
		/* Avoid duplication of identical images. */
		if (IS_UP(n) && IS_DOWN(n) && IS_LEFT(n) && IS_RIGHT(n))
			images[next][n] = images[next][BITMAPS - 1];
		else if (IS_LEFT_UP(n) && (IS_LEFT(n) || IS_UP(n)))
			images[next][n] = images[next][n & ~LEFT_UP];
		else if (IS_LEFT_DOWN(n) && (IS_LEFT(n) || IS_DOWN(n)))
			images[next][n] = images[next][n & ~LEFT_DOWN];
		else if (IS_RIGHT_UP(n) && (IS_RIGHT(n) || IS_UP(n)))
			images[next][n] = images[next][n & ~RIGHT_UP];
		else if (IS_RIGHT_DOWN(n) && (IS_RIGHT(n) || IS_DOWN(n)))
			images[next][n] = images[next][n & ~RIGHT_DOWN];
		else if (image_needed(n)) {
			if (!create_an_image(
#ifndef WINVER
					display, visual,
#endif
					n, use3D, next, size))
				return 0;
		}
	}
	return 1;
}

#ifdef WINVER
#ifndef None
#define None NULL
#endif
#endif
void
free_images(int next)
{
	int n;

	if (images[next][BITMAPS - 1] != None) {
#ifdef WINVER
		DeleteObject(images[next][BITMAPS - 1]);
#else
		(void) XDestroyImage(images[next][BITMAPS - 1]);
#endif
		images[next][BITMAPS - 1] = None;
	}
	for (n = 0; n < BITMAPS - 1; n++) {
		/* Don't bother to free duplicates */
		if (IS_UP(n) && IS_DOWN(n) && IS_LEFT(n) && IS_RIGHT(n))
			images[next][n] = None;
		else if (IS_LEFT_UP(n) && (IS_LEFT(n) || IS_UP(n)))
			images[next][n] = None;
		else if (IS_LEFT_DOWN(n) && (IS_LEFT(n) || IS_DOWN(n)))
			images[next][n] = None;
		else if (IS_RIGHT_UP(n) && (IS_RIGHT(n) || IS_UP(n)))
			images[next][n] = None;
		else if (IS_RIGHT_DOWN(n) && (IS_RIGHT(n) || IS_DOWN(n)))
			images[next][n] = None;
		else if (images[next][n] != None) {
#ifdef WINVER
			DeleteObject(images[next][n]);
#else
			(void) XDestroyImage(images[next][n]);
#endif
			images[next][n] = None;
		}
	}
}

#ifdef IMAGE
int
polyPixmap(int i, int j, int p[6][6], int s)
{
	unsigned int pmid = 0;

	if (ARR(i,j,p,s) != ARR(i-1,j,p,s))
		pmid |= LEFT;
	if (ARR(i,j,p,s) != ARR(i,j+1,p,s))
		pmid |= DOWN;
	if (ARR(i,j,p,s) != ARR(i+1,j,p,s))
		pmid |= RIGHT;
	if (ARR(i,j,p,s) != ARR(i,j-1,p,s))
		pmid |= UP;
	if (ARR(i,j,p,s) != ARR(i-1,j+1,p,s))
		pmid |= LEFT_DOWN;
	if (ARR(i,j,p,s) != ARR(i+1,j+1,p,s))
		pmid |= RIGHT_DOWN;
	if (ARR(i,j,p,s) != ARR(i+1,j-1,p,s))
		pmid |= RIGHT_UP;
	if (ARR(i,j,p,s) != ARR(i-1,j-1,p,s))
		pmid |= LEFT_UP;
	return pmid;
}
#endif
